extern crate rltk;
use rltk::{Point, Rltk, VirtualKeyCode};
extern crate specs;
use super::{CombatStats, Map, Player, Position, RunState, State, Viewshed, WantsToMelee};
use specs::prelude::*;
use std::cmp::{max, min};


fn clamp<T: PartialOrd>(low: T, high: T, val: T) -> T {
    if val < low {
        low
    } else if val > high {
        high
    } else {
        val
    }
}

#[derive(Component)]
pub enum GameEvent {
    MoveTo {
        destination: usize,
        tgt_x: i32,
        tgt_y: i32,
    },
    Attack { target: Entity },
}

pub struct HandleGameEvent;

impl<'a> System<'a> for HandleGameEvent {
    type SystemData = (
        Entities<'a>,
        WriteStorage<'a, GameEvent>,
        WriteStorage<'a, WantsToMelee>,
        WriteStorage<'a, Position>,
        WriteStorage<'a, Viewshed>,
        WriteExpect<'a, Point>,
        ReadExpect<'a, Map>,
    );

    fn run(&mut self, (entities, mut game_events, mut melee, mut pos, mut viewshed, mut point, map): Self::SystemData) {
        for (e, event) in (&entities, &game_events).join() {
            match event {
                GameEvent::Attack { target } => {
                    melee.insert(e, WantsToMelee { target: *target })
                        .expect("Unable to add melee intention for player");
                }
                GameEvent::MoveTo { destination, tgt_x, tgt_y } => {
                    if !map.blocked[*destination] {
                        viewshed.get_mut(e).expect("Unable to find viewshed").dirty = true;
                        pos.get_mut(e).expect("Unable to find position").x = *tgt_x;
                        pos.get_mut(e).expect("Unable to find position").y = *tgt_y;
                        point.x = clamp(0, 79, *tgt_x);
                        point.y = clamp(0, 49, *tgt_y);
                    }
                }
            }
        }
        game_events.clear();
    }
}

pub struct HandleInputEvent;

impl<'a> System<'a> for HandleInputEvent {
    type SystemData = (
        WriteStorage<'a, InputEvent>,
        ReadExpect<'a, Map>,
        Entities<'a>,
        ReadStorage<'a, Position>,
        ReadStorage<'a, CombatStats>,
        WriteStorage<'a, GameEvent>,
    );

    fn run(&mut self, (mut event, map, entities, position, stats, mut game_events): Self::SystemData) {
        for (e, ev, pos) in (&entities, &event, &position).join() {
            match *ev {
                InputEvent::PlayerMovement { delta_x, delta_y } => {
                    // depending on what's in the target cell, we
                    // either want to attack or move
                    let tgt_x = pos.x + delta_x;
                    let tgt_y = pos.y + delta_y;
                    let destination = map.xy_idx(tgt_x, tgt_y);
                    if let Some(target) = map.get_attackable(destination, &stats) {
                        game_events.insert(e, GameEvent::Attack { target })
                            .expect("Unable to add GameEvent");
                    } else {
                        game_events.insert(e, GameEvent::MoveTo { destination, tgt_x, tgt_y })
                            .expect("Unable to add GameEvent");
                    }
                }
            }
        }
        event.clear();
    }
}


/// This type represents all the "low-level" actions we have available
/// to us, corresponding directly to the set of actions we have bound
/// on the keys
#[derive(Component)]
pub enum InputEvent {
    // attempt to move the player some amount in the given direction
    // (delta_x and delta_y should be -1, 0, or 1)
    PlayerMovement {
        delta_x: i32,
        delta_y: i32,
    },
}

pub fn player_input(gs: &mut State, ctx: &mut Rltk) -> RunState {
    let player = *gs.ecs.fetch::<Entity>();
    let mut input = gs.ecs.write_storage::<InputEvent>();
    // Player movement
    match ctx.key {
        None => return RunState::AwaitingInput, // Nothing happened
        Some(key) => match key {
            VirtualKeyCode::Left | VirtualKeyCode::Numpad4 | VirtualKeyCode::H =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: -1, delta_y: 0 }),

            VirtualKeyCode::Right | VirtualKeyCode::Numpad6 | VirtualKeyCode::L =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: 1, delta_y: 0 }),

            VirtualKeyCode::Up | VirtualKeyCode::Numpad8 | VirtualKeyCode::K =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: 0, delta_y: -1 }),

            VirtualKeyCode::Down | VirtualKeyCode::Numpad2 | VirtualKeyCode::J =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: 0, delta_y: 1 }),

            // Diagonals
            VirtualKeyCode::Numpad9 | VirtualKeyCode::U =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: 1, delta_y: -1 }),

            VirtualKeyCode::Numpad7 | VirtualKeyCode::Y =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: -1, delta_y: -1 }),

            VirtualKeyCode::Numpad3 | VirtualKeyCode::N =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: 1, delta_y: 1 }),

            VirtualKeyCode::Numpad1 | VirtualKeyCode::B =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: -1, delta_y: 1 }),

            _ => return RunState::AwaitingInput,
        },
    }.expect("Unable to insert player movement");
    RunState::PlayerTurn
}
